import * as d3 from 'd3';
import ELK from 'elkjs';


import { Module } from './layout';
import { globalLookup, globalSetting } from '../global';
import { PortRender } from './port';
import { InstantiationRender } from './instantiation';
import { CellRender } from './cell';
import { ConnectionRender } from './connection';
import { WireRender } from './wire';

export class NetlistRender {
    /**
     * 
     * @param {YosysRawNet} rawNet 
     * @param {number} renderHeight
     * @param {number} renderWidth
     */
    constructor() {
        /**
         * @type {ElkGraph}
         */
        this.elkGraph = {
            id: 'root',
            layoutOptions: {
                'elk.progressBar': true,
                'layered.nodePlacement.strategy': 'SIDE_BASED'
            },
            children: [],
            edges: []
        };

        /**
         * @type {Map<string, Module>}
         */
        this.nameToModule = new Map();

    }

    /**
     * @description 加载 yosys 格式的 json
     * @param {YosysRawNet} rawNet 
     */
    load(rawNet) {
        /**
         * @type {YosysRawNet}
         */
        this.rawNet = rawNet;

        // 转换为 elkjs 格式的 graph
        for (const [moduleName, rawModule] of Object.entries(rawNet.modules)) {
            const top = parseInt(rawModule.attributes.top);
            // 一开始只渲染 top 模块
            if (top) {
                const module = new Module(moduleName, rawModule);
                const portNodes = module.makeNetsElkNodes();
                const cellNodes = module.makeCellsElkNodes();
                const [constantNodes, connectionEdges] = module.makeConnectionElkNodes();

                this.elkGraph.children.push(...portNodes);
                this.elkGraph.children.push(...cellNodes);
                this.elkGraph.children.push(...constantNodes);
                this.elkGraph.edges.push(...connectionEdges);

                this.nameToModule.set(moduleName, module);
            }
        }
    }

    /**
     * @description 根据信息创建布局对象
     * @returns {Promise<ElkNode>}
     */
    async createLayout() {
        const elk = new ELK();
        const graph = this.elkGraph;
        const layoutGraph = await elk.layout(graph, {
            layoutOptions: {
                'org.eclipse.elk.layered.spacing.nodeNodeBetweenLayers': 35,
                'org.eclipse.elk.spacing.nodeNode': 35,
                'org.eclipse.elk.layered.layering.strategy': 'LONGEST_PATH'
            }
        });

        return layoutGraph;
    }

    /**
     * @description 窗体大小发生变化时
     */
    registerResizeHandler(element) {
        this.resizeMonitor = new ResizeObserver(entries => {
            for (const entry of entries) {
                const { height, width } = entry.contentRect;
                this.renderHeight = height;
                this.renderWidth = width;
                if (this.selection) {
                    this.selection
                        .attr('width', this.renderWidth)
                        .attr('height', this.renderHeight)
                }
            }
        });
        this.resizeMonitor.observe(element);
    }

    /**
     * 
     * @param {ElkNode} computedLayout 
     * @param {string} container 类似于 "#chart"
     * @returns {d3.Selection}
     */
    async render(computedLayout, container) {
        const virtualHeight = computedLayout.height;
        const virtualWidth = computedLayout.width;

        const element = document.querySelector(container);
        this.renderHeight = element.clientHeight;
        this.renderWidth = element.clientWidth;
        this.registerResizeHandler(element);

        // 根据 height 进行放缩（可以通过设置进行调整）
        const ratio = this.renderHeight / virtualHeight;

        // 遍历计算布局进行创建
        const svg = d3.select(container)
            .selectAll('svg')
            .attr('width', this.renderWidth)
            .attr('height', this.renderHeight);

        // 将分组作为一个后续操作的 parent selection
        const g = svg.append('g');

        await this.renderLine(g, computedLayout, ratio);
        await this.renderEntity(g, computedLayout, ratio);

        // svg 挂载为全局注册的 selection
        this.selection = svg;

        // 注册平移和缩放
        // this.registerRenderTransform(g);
        
        return svg;
    }

    /**
     * @description 绘制实体
     * @param {d3.Selection} parentSelection 
     * @param {ElkNode} computedLayout 
     * @param {number} ratio
     */
    async renderEntity(parentSelection, computedLayout, ratio) {
        // node 可能是如下的几类
        // - module 的 port
        // - 器件（基础器件 & 例化模块）
        // - 器件的 port
        const skinManager = globalLookup.skinManager;

        // 创建各个主要实体的 render
        this.cellRender = new CellRender(parentSelection);
        this.portRender = new PortRender(parentSelection);
        this.instantiationRender = new InstantiationRender(parentSelection);
        this.connectionRender = new ConnectionRender(parentSelection);

        for (const node of computedLayout.children) {
            const skin = skinManager.querySkin(node.renderName);
            if (skin) {
                // 具有 skin 的器件
                this.cellRender.addAsD3DataItem(node, skin.meta.svgDoc.documentElement);
            } else {
                if (node.renderType === 'port') {
                    this.portRender.addAsD3DataItem(node);
                } else {
                    // 没有 skin 的器件或者端口
                    this.instantiationRender.addAsD3DataItem(node);
                }
            }

            // 如果存在 port，绘制 port
            for (const cellPort of node.ports || []) {
                this.connectionRender.addAsD3DataItem(cellPort, node); 
            }
        }

        this.portRender.render();
        // TODO: 实现它
        // this.instantiationRender.render();
        this.cellRender.render();
        this.connectionRender.render();        
    }

    /**
     * @description 绘制连线
     * @param {d3.Selection} parentSelection 
     * @param {ElkNode} computedLayout 
     * @param {number} ratio
     */
    async renderLine(parentSelection, computedLayout, ratio) {

        this.wireRender = new WireRender(parentSelection);

        for (const edge of computedLayout.edges) {
            for (const section of edge.sections || []) {
                const points = [];
                points.push(section.startPoint);
                for (const point of section.bendPoints || []) {
                    points.push(point);
                }
                points.push(section.endPoint);
                this.wireRender.addAsD3DataItems(points);
            }
        }

        this.wireRender.render();
    }

    /**
     * @description 注册平移和缩放
     * @param {d3.Selection} parentSelection 
     */
    registerRenderTransform(parentSelection) {
        // 创建缩放行为
        const zoom = d3.zoom()
            // 设置缩放范围（最小缩放比例 0.5，最大缩放比例 5）
            .scaleExtent([0.5, 5])
            .on("zoom", zoomed);

        // 将缩放行为应用到 SVG
        this.selection.call(zoom);

        // 缩放事件处理函数
        function zoomed(event) {
            const { transform, sourceEvent } = event;            

            if (sourceEvent && sourceEvent.type === "wheel") {
                if (globalSetting.renderAnimation) {
                    parentSelection
                        .transition()
                        .duration(350)
                        .attr('transform', transform);
                } else {
                    parentSelection.attr('transform', transform);
                }
            } else if (sourceEvent && sourceEvent.type === "mousemove") {
                if (globalSetting.renderAnimation) {
                    parentSelection.attr('transform', transform);
                } else {
                    parentSelection.attr('transform', transform);
                }
            }
        }
    }

    /**
     * @description 从 globalLookup 中更新 svg 的方位
     */
    updateLocationFromGlobal() {
        // const svg = globalLookup.netlistRender.selection;
        // if (!svg) {
        //     return;
        // }
        
        // svg.attr('transform', `translate(${globalLookup.svgTranslateX}, ${globalLookup.svgTranslateY}) scale(${globalLookup.svgScale})`);    
    }

    /**
     * @description 渲染 module 中的 port
     */
    renderPorts() {

    }

    /**
     * @description 渲染 module 中的例化模块
     */
    renderInstantiations() {

    }

    /**
     * @description 渲染 module 中的基础器件
     */
    renderCells() {

    }

    /**
     * @description 渲染每一个 例化模块/基础器件 的接入端口
     */
    renderConnections() {

    }

    /**
     * @description 渲染连线
     */
    renderWires() {

    }
}